import os
import subprocess
import falcon
import json
from tveapi.common import resource as res
from oslo_log import log
import requests
import re
from urllib.parse import urlparse
from tveapi.json_tools import json_tools
from jsonschema import validate
from tveapi.common import json_schemas
from tveapi.common.utils import decrypt_manager_secret_str

LOG = log.getLogger(__name__)

name_convention = re.compile("^[a-z0-9_][a-z0-9\/_.-]{0,127}$")


def get_available_space():
    """
    This function will return available space on controller in MB.
    """
    disk = os.statvfs('/')
    available_space = (disk.f_bavail * disk.f_frsize) / 1024 ** 2
    return available_space


def validate_vnf_status_file():
    """
    This function validates /kubernetes/vnfs/vnfs.json status file.
    """
    try:
        if not os.path.exists(res.K8S_VNFS_DIR):
            subprocess.call(
                "sudo chmod a+w /kubernetes/vnfs/ || true", shell=True)
            subprocess.call(
                f'export KUBECONFIG=/home/tveapi/.kube/config; sudo chmod a+w {res.K8S_VNFS_DIR}>>{res.K8S_VNFS_DIR}', shell=True)
        with open(res.K8S_VNFS_DIR) as fd:
            try:
                validate(instance=json.loads(fd.read()), schema=json_schemas.kubernetes_vnf_images)
            except Exception as e:
                images = {'images': []}
                for filename in os.listdir('/kubernetes/vnfs'):
                    if not ':' in filename:
                        continue
                    name = filename.split(':')[0]
                    tag = filename.split(':')[1]
                    vnf_file = f'{name}:{tag}'
                    images['images'].append({'name': name, 'tag': tag, 'file': vnf_file})
                json_tools.write_json_file(images, res.K8S_VNFS_DIR)
        fd.close()
    except Exception as e:
        pass


class Image(object):
    """
    OS_USERNAME, OS_PASSWORD for the internal system user. See local-openrc.env
    Get TOKEN : TOKEN=$(curl -s -k -d '{"user":"OS_USERNAME","password":"OS_PASSWORD"}' -H "Content-type: application/json" -X GET https://127.0.0.1:17374/v3/auth/tokens)

    API tests:
    list all images : curl -s -H "X-Auth-Token: $TOKEN" --insecure -X GET  https://127.0.0.1:7373/v1/kubernetes/image

    get status of pulling for specific image : curl -s -H "X-Auth-Token: $TOKEN" -k --insecure -d '{"name": "cvedia/public", "tag": "runtime-x86_64-public-2022.11.1"}' -X GET  https://127.0.0.1:7373/v1/kubernetes/image

    import image from remote repository: curl -s -H "X-Auth-Token: $TOKEN" --insecure -F'jsonData={"name":"cvedia/public","tag":"runtime-x86_64-public-2022.11.1"}' -X PUT  https://127.0.0.1:7373/v1/kubernetes/image/import
    import image from public repository : curl -s -H "X-Auth-Token: $TOKEN" --insecure -F'jsonData={"url":"https://docker.io","name":"mysql","tag":"latest"}' -X PUT  https://127.0.0.1:7373/v1/kubernetes/image/import

    curl -s -H "X-Auth-Token: $TOKEN" --insecure -F'jsonData={"url":"https://mcr.microsoft.com","name":"dotnet/samples","tag":"dotnetapp-alpine-slim-amd64"}' -X PUT  https://127.0.0.1:7373/v1/kubernetes/image/import

    Authentication for public repository : curl -s -H "X-Auth-Token: $TOKEN" --insecure -F'jsonData={"url":"https://docker.io","name":"mysql","tag":"latest", "username":"telco", "password":"bjhv"}' -X PUT https://127.0.0.1:7373/v1/kubernetes/image/import

    upload .tar image : curl -s -H "X-Auth-Token: $TOKEN" --insecure -F"fileName=@cirros.tar" -F"jsonData={\"name\":\"cirrosvs\",\"tag\":\"2050\"}" -X PUT  https://127.0.0.1:7373/v1/kubernetes/image/upload

    upload .qcow2 image : curl -s -H "X-Auth-Token: $TOKEN" --insecure -F"fileName=@Fedora-Cloud-Base-36-1.5.x86_64.qcow2" -F"jsonData={\"name\":\"fedoravs\",\"tag\":\"111\"}" -X PUT  https://127.0.0.1:7373/v1/kubernetes/image/upload

    download images from public repo for ftp : curl -s -H "X-Auth-Token: $TOKEN" --insecure -H "Content-Type: application/json" -d '{"url": "ftp://anonymous:@10.3.112.14:21/VNFs/pfsense/pfsense_4.2.qcow2", "name": "test", "tag": "123"}' -X POST https://127.0.0.1:7373/v1/kubernetes/image/download

    HTTP/HTTPS : curl -s -H "X-Auth-Token: $TOKEN" --insecure -H "Content-Type: application/json" -d '{"url": "http://10.3.112.14/VNFs/pfsense/pfsense_4.2.qcow2", "name": "test", "tag": "123"}' -X POST https://127.0.0.1:7373/v1/kubernetes/image/download
    
    delete images : curl -s -H "X-Auth-Token: $TOKEN" --insecure -d '{"delete":[{"name": "cirrosvs", "tag": "2027"}]}' -X PUT  https://127.0.0.1:7373/v1/kubernetes/image/delete
    """

    class GetImages(res.BaseResource):
        def on_get(self, req, resp):
            images = {'images': {}}

            # Get status of pulling image
            if req.content_length is not None:
                json_req = self.json_body(req)
                status_file = '/tmp/registry_api_{}'.format(
                    re.sub('[/:]', '_', json_req['name']+':'+json_req['tag']))
                if os.path.exists(status_file):
                    with open(status_file, 'r') as f:
                        status = f.read()
                    resp.body = {'import': status}
                    resp.status = falcon.HTTP_200
                    return

            try:
                used_images = ""
                try:
                    used_images = subprocess.check_output(
                        "export KUBECONFIG=/home/tveapi/.kube/config; kubectl get pods -A -o jsonpath=\"{.items[*].spec.containers[*].image}\"", shell=True).decode()
                    used_images += subprocess.check_output(
                        "export KUBECONFIG=/home/tveapi/.kube/config; kubectl get vm -A -o jsonpath=\"{.items[*].spec.dataVolumeTemplates[*].spec.source.http.url}\" 2> /dev/null", shell=True).decode().replace('http://cdi-fileserver.cdi.svc.cluster.local:8080/', '')
                except Exception as e:
                    pass
                get_repos = 'curl http://localhost:5001/v2/_catalog'
                repos = json.loads(
                    subprocess.check_output(get_repos, shell=True))
                if repos:
                    for repo in repos['repositories']:
                        get_tags = 'curl http://localhost:5001/v2/%s/tags/list' % repo
                        tags = json.loads(
                            subprocess.check_output(get_tags, shell=True))
                        if tags['tags'] == None:
                            continue
                        images['images'][repo] = []
                        for tag in tags['tags']:
                            image = ':5001/'+repo+':'+tag
                            if image in used_images:
                                images['images'][repo].append(
                                    {'tag': tag, 'in_use': True})
                            else:
                                images['images'][repo].append(
                                    {'tag': tag, 'in_use': False})
                validate_vnf_status_file()
                vnf_data = json_tools.get_json_dict_from_file(res.K8S_VNFS_DIR)
                for image in vnf_data['images']:
                    image_name = image['name']
                    image_tag = image['tag']
                    image_file = f"{image['name']}:{image['tag']}"
                    if not image_name in images['images']:
                        images['images'][image_name] = []
                    if os.path.exists(f'/kubernetes/vnfs/{image_file}'):
                        if image_file in used_images:
                            images['images'][image_name].append({'tag': image_tag, 'in_use': True})
                        else:
                            images['images'][image_name].append({'tag': image_tag, 'in_use': False})
            except Exception as e:
                self.on_fail(resp, e)
                return
            resp.status = falcon.HTTP_200
            resp.body = json.dumps(images, indent=4)

    @falcon.before(res.BaseResource.file_size(limit=2048 * 1024 ** 2))
    class Upload(res.BaseResource):
        def on_put(self, req, resp):
            input_file = req.get_param('fileName')
            filename = input_file.filename.replace(" ", "")
            image_data = req.get_param('jsonData')
            data = input_file.file.read()
            fsize = int(len(data)) / 1024 ** 2

            try:
                # Check for correct image name and tag
                if image_data != 2:
                    json_req = json.loads(image_data)
                    if 'name' not in json_req or 'tag' not in json_req or (json_req['name'] and json_req['tag']) == '':
                        self.on_fail(resp, 'Name or tag in jsonData missing.')
                        return
                    elif not name_convention.match(json_req['name']) or not name_convention.match(json_req['tag']):
                        self.on_fail(
                            resp, 'Name and tag components may contain only lowercase letters, digits, separators and slashes.')
                        return
                else:
                    self.on_fail(resp, 'Name or tag in jsonData missing.')
                    return

                if filename:
                    # Check for free space on disk after accept new data
                    if fsize <= get_available_space() - fsize * 2:
                        file_path = os.path.join(res.TEMPORARY_DIR, filename)
                        temp_file_path = file_path + '~'

                        open(temp_file_path, 'wb').write(data)
                        os.rename(temp_file_path, file_path)
                    else:
                        self.on_fail(resp, 'No available space on device.')
                        return
                else:
                    self.on_fail(resp, 'File not found.')
                    return
            except OSError as e:
                LOG.error('fail exception OSError')
                self.on_fail(resp, e)
                return
            except Exception as e:
                self.on_fail(resp, e)
                return

            validate_vnf_status_file()
            subprocess.Popen('sudo python3 /scripts/registry_api.py {} {} \'{}\''.format(json_req['name'], json_req['tag'], filename), shell=True, close_fds=False)

            resp.body = {'import': 'starting'}
            resp.status = falcon.HTTP_202

    class Import(res.BaseResource):
        def on_put(self, req, resp):
            image_data = req.get_param('jsonData')

            try:
                # Check for correct image name and tag
                if image_data != 2:
                    json_req = json.loads(image_data)
                    if not ('name' in json_req and 'tag' in json_req and json_req['name'] and json_req['tag']):
                        self.on_fail(resp, 'Name or tag in jsonData missing.')
                        return
                    elif not name_convention.match(json_req['name']) or not name_convention.match(json_req['tag']):
                        self.on_fail(
                            resp, 'Name and tag components may contain only lowercase letters, digits, separators and slashes.')
                        return
                else:
                    self.on_fail(resp, 'Name or tag in jsonData missing.')
                    return

                # Because to login in docker.io, user can use upper case letters, but ctr command is case sensitive.
                # If your username is tester, you can login into docker.io with TeStEr username (valid only for docker.io).
                username = json_req.get('username')
                password = decrypt_manager_secret_str(json_req.get('password'), res.FERNET_KEY)
                
                registry_url = json_req.get('url', 'https://hub.docker.com')

                if "docker" in registry_url or not registry_url:
                    search_path = 'library/' + json_req['name']
                    if '/' in json_req['name']:
                        search_path = json_req['name']

                    # Check download rate limit (100 per anonymous user and 200 per authenticated user on every 6 hours)
                    # https://docs.docker.com/docker-hub/download-rate-limit/
                    if username and password:
                        username = username.lower()
                        token_url = 'https://auth.docker.io/token?service=registry.docker.io&scope=repository:{}:pull&account={}'.format(search_path, username)
                        auth = (username, password)
                    else:
                        token_url = 'https://auth.docker.io/token?service=registry.docker.io&scope=repository:{}:pull'.format(search_path)
                        auth = None

                    token = requests.get(token_url, auth=auth)

                    headers = {"Authorization": "Bearer {}".format(token.json()['token'])}

                    tag = requests.get(
                        'https://registry-1.docker.io/v2/{}/manifests/{}'.format(search_path, json_req['tag']), headers=headers)

                    if tag.status_code == 429:
                        self.on_fail(
                                resp, 'You have reached the max number of Docker image downloads (100 downloads per 6 hours and 200 for authenticated). Try with another authenticated user.')
                        return
                    elif tag.status_code != 200:
                        self.on_fail(
                            resp, 'Unable to find image with this name and tag.')
                        return

                    # Check if image exists in remote repository and return size in MB
                    # For images on Docker Hub
                    # default arch = amd64
                    # TODO: need to recheck
                    image_size = 0
                    try:
                        for image in json.loads(tag.text)['images']:
                            if image['architecture'] == 'amd64':
                                image_size = image['size']

                                # Convert to MB
                                image_size /= 1048576
                                print(f'image_size: {image_size}')

                                # Removed this, because some of tags are without layers info
                                # if image_size == 0:
                                #     self.on_fail(
                                #         resp, 'Unable to find image with this name and tag.')
                                #     return
                                if image_size >= get_available_space() - image_size * 2:
                                    self.on_fail(resp, 'No available space on device.')
                                    return
                    except Exception as e:
                        pass
                    
                else:
                    if "mcr.microsoft.com" in registry_url:
                        headers = {}
                    else:
                        # If another registry requires authentication, add logic here
                        self.on_fail(resp, 'Unsupported registry.')
                        return

                    tag_url = "{}/v2/{}/manifests/{}".format(registry_url.rstrip('/'), json_req['name'], json_req['tag'])
                    tag = requests.get(tag_url, headers=headers)

                    if tag.status_code != 200:
                        self.on_fail(resp, 'Unable to find image with this name and tag.')
                        return
                    else:
                        json_data = json.loads(tag.text)

                        if 'architecture' in json_data and json_data['architecture'] == 'amd64':
                            pass
                        else:
                            self.on_fail(
                                resp, 'Unable to find image with this name and tag.')
                            return
                            #image_size = tag.text['size']
                    # Convert to MB
                    #image_size /= 1048576

                    #if image_size == 0:
                    #    self.on_fail(
                    #        resp, 'Unable to find image with this name and tag.')
                    #    return
                    #elif image_size >= get_available_space() - image_size * 2:
                    #    self.on_fail(resp, 'No available space on device.')
                    #    returns

                validate_vnf_status_file()
                if 'url' in json_req and json_req['url']:
                    parsed_url = urlparse(json_req['url'])
                    registry_domain = parsed_url.netloc if parsed_url.scheme == 'https' else parsed_url.path
                   
                    if username is None and password is None:
                        password=""
                        username=""
                    
                    subprocess.Popen('sudo python3 /scripts/registry_api.py {} {} {} {} {}'.format(
                        json_req['name'], json_req['tag'], registry_domain, username, password), shell=True, close_fds=False)
                else:
                    subprocess.Popen('sudo python3 /scripts/registry_api.py {} {} None {} {}'.format(
                        json_req['name'], json_req['tag'], username, password), shell=True, close_fds=False)
                
            #except subprocess.CalledProcessError:
            #    self.on_fail(
            #            resp, 'Unable to find image with this name and tag.')
            #    return
            except Exception as e:
                self.on_fail(resp, e)
                return

            resp.body = {'import': 'starting'}
            resp.status = falcon.HTTP_202

    
    
    class Download(res.BaseResource):
        def on_post(self, req, resp):
            try:
                json_req = self.json_body(req)

                # Check for required parameters
                required_params = ['url', 'name', 'tag']
                for param in required_params:
                    if param not in json_req:
                        self.on_fail(resp, 'Required parameter missing: {}'.format(param))
                        return

                # Upload the downloaded file to the registry
                validate_vnf_status_file()
                process = subprocess.Popen(
                    'sudo python3 /scripts/registry_api.py {} {} "{}"'.format(
                    json_req['name'], json_req['tag'], json_req['url']),
                    shell=True, close_fds=False) 
                
                resp.body = {'download': 'starting', 'upload': 'starting'}
                resp.status = falcon.HTTP_200
            except Exception as e:
                self.on_fail(resp, e)
                return

    class Delete(res.BaseResource):
        def on_put(self, req, resp):
            json_req = self.json_body(req)
            images = json_req['delete']
            response = {}
            url = 'http://localhost:5001/v2/{}/manifests/{}'
            headers = ['application/vnd.docker.distribution.manifest.v2+json', 'application/vnd.oci.image.manifest.v1+json']

            if images == "":
                self.on_fail(resp, "Please, add image name.")
                return
            try:
                used_images = ""
                try:
                    used_images = subprocess.check_output(
                        "export KUBECONFIG=/home/tveapi/.kube/config; kubectl get pods -A -o jsonpath=\"{.items[*].spec.containers[*].image}\"", shell=True).decode()
                    used_images += subprocess.check_output(
                        "export KUBECONFIG=/home/tveapi/.kube/config; kubectl get vm -A -o jsonpath=\"{.items[*].spec.dataVolumeTemplates[*].spec.source.http.url}\" 2> /dev/null", shell=True).decode().replace('http://cdi-fileserver.cdi.svc.cluster.local:8080/', '')
                except Exception as e:
                    pass
                for image in images:
                    # Check for correct image name and tag
                    if 'name' not in image or 'tag' not in image or (image['name'] and image['tag']) == '':
                        self.on_fail(resp, 'Name or tag in data is missing.')
                        return
                    elif not name_convention.match(image['name']) or not name_convention.match(image['tag']):
                        self.on_fail(
                            resp, 'Name and tag components may contain only lowercase letters, digits, separators and slashes.')
                        return
                    # Check if image is used by POD/VM
                    image_name = image['name']+':'+image['tag']
                    current_image = ':5001/' + image_name
                    vnf_image = f"{image['name']}:{image['tag']}"
                    if current_image in used_images or vnf_image in used_images:
                        response[image_name] = [
                            {'delete': 'failed', 'description': 'Image is used by VM or pod.'}]
                        continue

                    # Delete status file
                    status_file = '/tmp/registry_api_{}'.format(
                        re.sub('[/:]', '_', image_name))
                    if os.path.exists(status_file):
                        os.system("sudo rm {}".format(status_file))

                    for manifest in headers:
                        digest = requests.get(url.format(image['name'], image['tag']),
                                            headers={'Accept': f'{manifest}'},
                                            verify=False, timeout=(5, 5))
                        if digest.status_code == 200:
                            break
                    if digest.status_code == 200:
                        requests.delete(url.format(
                            image['name'], digest.headers['Docker-Content-Digest']))
                        # Run garbage-collection (https://docs.docker.com/registry/garbage-collection/)
                        subprocess.call(
                            'export KUBECONFIG=/home/tveapi/.kube/config; sudo /usr/bin/registry garbage-collect -m /kubernetes/templates/registry/registry-config.yaml --delete-untagged', shell=True)
                        # here delete /var/lib/local-registry/docker/registry/v2/repositories/<name>
                        get_tags = 'curl http://localhost:5001/v2/%s/tags/list' % image['name']
                        tags = json.loads(
                            subprocess.check_output(get_tags, shell=True))
                        if tags['tags'] == None:
                            subprocess.call(
                            "export KUBECONFIG=/home/tveapi/.kube/config; sudo rm -rf /var/lib/local-registry/docker/registry/v2/repositories/%s" % image['name'], shell=True)
                        # PLEASE NOTE in registry version 2.7.1->latest :- At this moment we have successfully deleted our image.
                        # But if we again push the same image to local registry, we will see a reply saying the image layers still
                        # exist. In order to avoid this we need to restart our docker registry.
                        subprocess.call("export KUBECONFIG=/home/tveapi/.kube/config; sudo systemctl restart local-registry", shell=True)
                        response[image_name] = [
                            {'delete': 'ok', 'description': ''}]
                    else:
                        validate_vnf_status_file()
                        vnf_data = json_tools.get_json_dict_from_file(res.K8S_VNFS_DIR)
                        if not vnf_data['images']:
                            response[image_name] = [
                                {'delete': 'failed', 'description': 'Image is missing!'}]
                        for vnf in vnf_data['images']:
                            if vnf['file'] == vnf_image:
                                vnf_data['images'].remove({'name': image['name'], 'tag': image['tag'], 'file': vnf_image})
                                json_tools.write_json_file(vnf_data, res.K8S_VNFS_DIR)
                                subprocess.call(
                                    'export KUBECONFIG=/home/tveapi/.kube/config; sudo rm -f /kubernetes/vnfs/%s' % vnf['file'], shell=True)
                                response[image_name] = [
                                    {'delete': 'ok', 'description': ''}]
                            else:
                                response[image_name] = [
                                    {'delete': 'failed', 'description': 'Image is missing!'}]
            except Exception as e:
                if "docker-content-digest" in str(e):
                    msg = 'Image is missing!'
                else:
                    msg = e
                response[image_name] = [
                        {'delete': 'failed', 'description': f'Error: {msg}'}]
                pass

            resp.body = json.dumps(response, indent=4)
            resp.status = falcon.HTTP_202
